use super::helpers;
use crate::{
    kv::{mdbx::*, MdbxWithDirHandle},
    models::*,
};
use async_trait::async_trait;
use ecomobicoin_jsonrpc::{types, ErigonApiServer};
use jsonrpsee::core::RpcResult;
use std::sync::Arc;

pub struct ErigonApiServerImpl<SE>
where
    SE: EnvironmentKind,
{
    pub db: Arc<MdbxWithDirHandle<SE>>,
}

#[async_trait]
impl<DB> ErigonApiServer for ErigonApiServerImpl<DB>
where
    DB: EnvironmentKind,
{
    async fn get_header_by_number(&self, block_number: u64) -> RpcResult<Option<types::Header>> {
        let db = self.db.clone();

        tokio::task::spawn_blocking(move || {
            let tx = db.begin()?;

            if let Some((block_number, _)) =
                helpers::resolve_block_id(&tx, types::BlockNumber::Number(block_number.into()))?
            {
                if let Some(BlockHeader {
                    parent_hash,
                    ommers_hash,
                    beneficiary,
                    state_root,
                    transactions_root,
                    behaviors_root,
                    receipts_root,
                    logs_bloom,
                    behaviors_bloom,
                    difficulty,
                    number,
                    gas_limit,
                    gas_used,
                    timestamp,
                    extra_data,
                    mix_hash,
                    nonce,
                    base_fee_per_gas,
                    behavior_total_quantity,
                    vdf_proof_challenge,
                    vdf_proof_result,
                    vdf_difficulty,
                }) = crate::accessors::chain::header::read(&tx, block_number)?
                {
                    return Ok(Some(types::Header {
                        parent_hash,
                        sha3_uncles: ommers_hash,
                        miner: beneficiary,
                        state_root,
                        transactions_root,
                        behaviors_root,
                        receipts_root,
                        logs_bloom,
                        behaviors_bloom,
                        difficulty,
                        number: number.0.into(),
                        gas_limit: gas_limit.into(),
                        gas_used: gas_used.into(),
                        timestamp: timestamp.into(),
                        extra_data: extra_data.into(),
                        mix_hash,
                        nonce,
                        base_fee_per_gas,
                        behavior_total_quantity: behavior_total_quantity.into(),
                        vdf_proof_challenge,
                        vdf_proof_result,
                        vdf_difficulty: vdf_difficulty.into(),
                    }));
                }
            }

            Ok(None)
        })
        .await
        .unwrap_or_else(helpers::joinerror_to_result)
    }
}
